from filecmp import cmp
from pathlib import Path

import tifffile

from brainglobe_segmentation.regions import IO as region_IO

regions_dir = Path.cwd() / "tests" / "data" / "regions"
VOXEL_SIZE = 100


def test_export_regions_to_file(tmp_path):
    image = tifffile.imread(regions_dir / "region.tiff")
    filename = tmp_path / "region.obj"
    region_IO.export_regions_to_file(image, filename, VOXEL_SIZE)

    cmp(regions_dir / "region.obj", tmp_path / "region.obj")



from pathlib import Path

import numpy as np

brainreg_dir = Path.cwd() / "tests" / "data" / "brainreg_output"

ATLAS_NAME = "example_mouse_100um"


def test_load_sample_space(segmentation_widget):
    segmentation_widget.atlas_space = False
    segmentation_widget.plugin = "brainglobe-napari-io.brainreg_read_dir"
    segmentation_widget.directory = brainreg_dir
    segmentation_widget.load_brainreg_directory()
    check_loaded_layers(segmentation_widget, 7)
    check_not_editable(segmentation_widget, atlas_space=False)


def test_load_atlas_space(segmentation_widget):
    segmentation_widget.atlas_space = True
    segmentation_widget.plugin = (
        "brainglobe-napari-io.brainreg_read_dir_atlas_space"
    )
    segmentation_widget.directory = brainreg_dir
    segmentation_widget.load_brainreg_directory()
    check_loaded_layers(segmentation_widget, 4)
    check_not_editable(segmentation_widget, atlas_space=True)


def test_layer_deletion(segmentation_widget):
    """
    Check that remove_layers() doesn't remove any layers that were present
    before brainglobe-segmentation added layers
    """
    assert len(segmentation_widget.viewer.layers) == 0
    segmentation_widget.viewer.add_points(np.array([[1, 1], [2, 2]]))
    assert len(segmentation_widget.viewer.layers) == 1
    segmentation_widget.atlas_space = False
    segmentation_widget.plugin = "brainglobe-napari-io.brainreg_read_dir"
    segmentation_widget.directory = brainreg_dir
    segmentation_widget.load_brainreg_directory()
    # Brainreg should load 7 new layers
    check_loaded_layers(segmentation_widget, 8)


def check_loaded_layers(widget, num_layers):
    assert len(widget.viewer.layers) == num_layers
    assert widget.base_layer.name == "Registered image"
    assert widget.atlas.atlas_name == "allen_mouse_50um"
    assert widget.metadata["orientation"] == "psl"
    assert widget.annotations_layer.name == widget.atlas.atlas_name


def check_not_editable(widget, atlas_space=False):
    assert widget.base_layer.editable is False
    assert widget.annotations_layer.editable is False
    if not atlas_space:
        assert widget.viewer.layers["Hemispheres"].editable is False


def test_general(segmentation_widget):
    segmentation_widget.atlas_space = True
    segmentation_widget.plugin = (
        "brainglobe-napari-io.brainreg_read_dir_atlas_space"
    )
    segmentation_widget.directory = brainreg_dir
    segmentation_widget.load_brainreg_directory()
    assert segmentation_widget.mean_voxel_size == 50
    check_defaults(segmentation_widget)
    check_paths(segmentation_widget)


def check_defaults(widget):
    assert widget.track_seg.point_size == int(100 / widget.mean_voxel_size)
    assert widget.track_seg.spline_size == int(50 / widget.mean_voxel_size)
    assert widget.track_seg.track_file_extension == ".points"
    assert widget.region_seg.image_file_extension == ".tiff"
    assert widget.region_seg.num_colors == 10
    assert widget.region_seg.brush_size == int(250 / widget.mean_voxel_size)
    assert widget.track_seg.spline_points_default == 1000
    assert widget.track_seg.spline_smoothing_default == 0.1
    assert widget.track_seg.fit_degree_default == 3
    assert widget.track_seg.summarise_track_default is True
    assert widget.region_seg.calculate_volumes_default is True
    assert widget.region_seg.summarise_volumes_default is True
    assert widget.boundaries_string == "Boundaries"


def check_paths(widget):
    assert widget.paths.main_directory == brainreg_dir / "segmentation"

    assert (
        widget.paths.segmentation_directory
        == brainreg_dir / "segmentation" / "atlas_space"
    )

    assert (
        widget.paths.regions_directory
        == brainreg_dir / "segmentation" / "atlas_space" / "regions"
    )

    assert (
        widget.paths.region_summary_csv
        == brainreg_dir
        / "segmentation"
        / "atlas_space"
        / "regions"
        / "summary.csv"
    )
    assert (
        widget.paths.tracks_directory
        == brainreg_dir / "segmentation" / "atlas_space" / "tracks"
    )



from pathlib import Path

import numpy as np
import pandas as pd
import pytest
import tifffile
from napari.layers import Labels

from brainglobe_segmentation.regions.analysis import (
    check_list_only_nones,
    summarise_brain_regions,
    summarise_single_brain_region,
)

region_image_path = Path(
    "tests/data/brainreg_output/segmentation/"
    "sample_space/regions/region_0.tiff"
)
atlas_resolution = (50, 50, 50)
label_name = "test_label"

# We run regression tests against a previously known "correct run"
# that produced these values
previous_image_space_values = {
    "region": "test_label",
    "area": 17842.00000,
    "bbox-0": 134,
    "bbox-1": 34,
    "bbox-5": 65,
    "centroid-2": 44.284385158614505,
}

previous_world_space_values = {
    "region": "test_label",
    "volume_mm3": 2.23025,
    "axis_0_min_um": 6700.0,
    "axis_1_min_um": 1700.0,
    "axis_2_min_um": 1250.0,
    "axis_0_max_um": 7850.0,
    "axis_1_max_um": 3350.0,
    "axis_2_max_um": 3250.0,
    "axis_0_center_um": 7250.0,
    "axis_1_center_um": 2483.154915368232,
    "axis_2_center_um": 2214.219257930725,
}


@pytest.fixture
def region_image():
    return tifffile.imread(region_image_path)


@pytest.fixture
def labels_layer(region_image):
    labels = Labels(region_image, name=label_name)
    return labels


@pytest.fixture
def empty_labels_layer(region_image):
    labels = Labels(np.zeros_like(region_image))
    return labels


def test_check_list_only_ones():
    assert check_list_only_nones([None, None, None]) is True
    assert check_list_only_nones([None, 1, None]) is False
    assert check_list_only_nones([1, "two", 3]) is False


def test_summarise_single_brain_region(labels_layer):
    """
    Test the summary of a single segmented region.
    N.B. these results are in "image" space, not "world" space,
    hence different to the other tests
    """
    df = summarise_single_brain_region(labels_layer)
    assert df.shape == (1, 11)
    for key, value in previous_image_space_values.items():
        assert df[key].iloc[0] == value


def test_summarise_brain_regions_all_not_empty(
    labels_layer, empty_labels_layer, tmp_path
):
    """
    Test summarise_brain_regions() when all the regions
    (usually segmented by the user)
    contain at least one labelled voxel

    N.B. these results are in "world" space, not "space" space,
    hence different to the test of a single region
    """
    filename = tmp_path / "summary.csv"

    summarise_brain_regions(
        [labels_layer, labels_layer, labels_layer], filename, atlas_resolution
    )
    df = pd.read_csv(filename)
    assert df.shape == (3, 11)
    for key, value in previous_world_space_values.items():
        assert df[key].iloc[0] == value


def test_summarise_brain_regions_some_empty(
    labels_layer, empty_labels_layer, tmp_path
):
    """
    Test summarise_brain_regions() when some of the regions
    (usually segmented by the user)
    contain no voxels (i.e. are empty arrays), but
    others contain labelled voxels.

    N.B. these results are in "world" space, not "space" space,
    hence different to the test of a single region
    """

    filename = tmp_path / "summary.csv"
    summarise_brain_regions(
        [empty_labels_layer, empty_labels_layer, labels_layer],
        filename,
        atlas_resolution,
    )
    df = pd.read_csv(filename)
    assert df.shape == (1, 11)
    for key, value in previous_world_space_values.items():
        assert df[key].iloc[0] == value


def test_summarise_brain_regions_all_empty(
    labels_layer, empty_labels_layer, tmp_path
):
    """
    Test summarise_brain_regions() when all the regions
    (usually segmented by the user)
    contain no voxels (i.e. are empty arrays)
    """
    filename = tmp_path / "summary.csv"

    # check correct behaviour if all regions are empty
    assert (
        summarise_brain_regions(
            [empty_labels_layer, empty_labels_layer],
            filename,
            atlas_resolution,
        )
        is None
    )



import shutil
from pathlib import Path

import pytest
from bg_atlasapi import BrainGlobeAtlas

from brainglobe_segmentation.segment import SegmentationWidget

atlas_name = "allen_mouse_50um"
brainreg_dir = Path.cwd() / "tests" / "data" / "brainreg_output"


@pytest.fixture
def allen_mouse_50um_atlas():
    return BrainGlobeAtlas(atlas_name)


@pytest.fixture
def segmentation_widget(make_napari_viewer):
    """
    Create a viewer, add the segmentation widget, and return the widget.
    The viewer can be accessed using ``widget.viewer``.
    """
    viewer = make_napari_viewer()
    widget = SegmentationWidget(viewer)
    viewer.window.add_dock_widget(widget)
    return widget


@pytest.fixture(scope="function")
def segmentation_widget_with_data_atlas_space(tmp_path, segmentation_widget):
    """
    Fixture to load a brainreg directory into the segmentation widget.
    Data is copied to tmpdir so that when it's loaded, so all the paths
    are set correctly.
    The segmentation data is then deleted so that saving/export
    can be properly tested
    """
    tmp_input_dir = tmp_path / "brainreg_output"
    shutil.copytree(brainreg_dir, tmp_input_dir)
    segmentation_widget.atlas_space = True
    segmentation_widget.plugin = (
        "brainglobe-napari-io.brainreg_read_dir_atlas_space"
    )
    segmentation_widget.directory = Path(tmp_input_dir)
    segmentation_widget.load_brainreg_directory()
    # delete segmentation data to ensure it's saved correctly in tests
    shutil.rmtree(segmentation_widget.paths.main_directory)
    return segmentation_widget



import numpy as np

from brainglobe_segmentation.atlas import utils as atlas_utils


def test_lateralise_atlas_image(allen_mouse_50um_atlas):
    atlas = allen_mouse_50um_atlas

    mask = np.random.random(atlas.annotation.shape) > 0.7
    masked_annotations = mask * atlas.annotation
    annotations_left, annotations_right = atlas_utils.lateralise_atlas_image(
        masked_annotations,
        atlas.hemispheres,
        left_hemisphere_value=atlas.left_hemisphere_value,
        right_hemisphere_value=atlas.right_hemisphere_value,
    )

    # not the best way to test this is functioning properly
    total_vals_in = np.prod(mask.shape)
    total_vals_out = len(annotations_left) + len(annotations_right)

    assert total_vals_in == total_vals_out



import numpy as np
from napari.layers import Labels

from brainglobe_segmentation.segmentation_panels.regions import (
    check_segmentation_in_correct_space,
)


def test_check_segmentation_in_correct_space():
    labels_layer_10 = Labels(np.zeros((10, 10, 10), dtype=int))
    labels_layer_15 = Labels(np.zeros((15, 15, 15), dtype=int))
    assert check_segmentation_in_correct_space(
        [labels_layer_10], labels_layer_10.data
    )
    assert not check_segmentation_in_correct_space(
        [labels_layer_10], labels_layer_15.data
    )



from qtpy.QtWidgets import (
    QCheckBox,
    QDoubleSpinBox,
    QLabel,
    QPushButton,
    QSpinBox,
)


def add_button(
    label,
    layout,
    connected_function,
    *,
    row: int = 0,
    column: int = 0,
    visibility=True,
    minimum_width=0,
    alignment="center",
    tooltip=None,
):
    button = QPushButton(label)
    if alignment == "center":
        pass
    elif alignment == "left":
        button.setStyleSheet("QPushButton { text-align: left; }")
    elif alignment == "right":
        button.setStyleSheet("QPushButton { text-align: right; }")

    button.setVisible(visibility)
    button.setMinimumWidth(minimum_width)

    if tooltip:
        button.setToolTip(tooltip)
    layout.addWidget(button, row, column)
    button.clicked.connect(connected_function)
    return button


def add_checkbox(
    layout, default, label, row: int = 0, column: int = 0, tooltip=None
):
    box = QCheckBox()
    box.setChecked(default)
    if tooltip:
        box.setToolTip(tooltip)
    layout.addWidget(QLabel(label), row, column)
    layout.addWidget(box, row, column + 1)
    return box


def add_float_box(
    layout,
    default,
    minimum,
    maximum,
    label,
    step,
    row: int = 0,
    column: int = 0,
    tooltip=None,
):
    box = QDoubleSpinBox()
    box.setMinimum(minimum)
    box.setMaximum(maximum)
    box.setValue(default)
    box.setSingleStep(step)
    if tooltip:
        box.setToolTip(tooltip)
    layout.addWidget(QLabel(label), row, column)
    layout.addWidget(box, row, column + 1)
    return box


def add_int_box(
    layout,
    default,
    minimum,
    maximum,
    label,
    row: int = 0,
    column: int = 0,
    tooltip=None,
):
    box = QSpinBox()
    box.setMinimum(minimum)
    box.setMaximum(maximum)
    # Not always set if not after min & max
    box.setValue(default)
    if tooltip:
        box.setToolTip(tooltip)
    layout.addWidget(QLabel(label), row, column)
    layout.addWidget(box, row, column + 1)
    return box



from filecmp import cmp
from pathlib import Path
from time import sleep

import numpy as np
import pandas as pd
import pytest
from tifffile import imread

brainreg_dir = Path.cwd() / "tests" / "data" / "brainreg_output"
validate_regions_dir = (
    brainreg_dir / "segmentation" / "atlas_space" / "regions"
)


@pytest.fixture
def test_regions_dir(tmp_path):
    tmp_input_dir = tmp_path / "brainreg_output"
    test_regions_dir = (
        tmp_input_dir / "segmentation" / "atlas_space" / "regions"
    )
    return test_regions_dir


def test_region_widget_layer_numbers(
    segmentation_widget_with_data_atlas_space, test_regions_dir
):
    assert len(segmentation_widget_with_data_atlas_space.viewer.layers) == 4
    assert len(segmentation_widget_with_data_atlas_space.label_layers) == 1


def test_add_new_region(
    segmentation_widget_with_data_atlas_space, test_regions_dir
):
    segmentation_widget_with_data_atlas_space.region_seg.add_new_region()
    assert len(segmentation_widget_with_data_atlas_space.viewer.layers) == 5
    assert len(segmentation_widget_with_data_atlas_space.label_layers) == 2
    assert (
        segmentation_widget_with_data_atlas_space.label_layers[0].name
        == "test_region"
    )
    assert (
        segmentation_widget_with_data_atlas_space.label_layers[1].name
        == "region_1"
    )


def test_add_existing_region(
    segmentation_widget_with_data_atlas_space, test_regions_dir
):
    test_layer = segmentation_widget_with_data_atlas_space.viewer.add_labels(
        segmentation_widget_with_data_atlas_space.label_layers[0].data,
        name="test_region_2",
    )
    segmentation_widget_with_data_atlas_space.viewer.layers.selection.select_only(
        test_layer
    )
    segmentation_widget_with_data_atlas_space.region_seg.add_region_from_existing_layer(
        override=True
    )
    assert len(segmentation_widget_with_data_atlas_space.viewer.layers) == 5
    assert len(segmentation_widget_with_data_atlas_space.label_layers) == 2


def test_region_analysis_without_save(
    segmentation_widget_with_data_atlas_space, test_regions_dir
):
    segmentation_widget_with_data_atlas_space.region_seg.calculate_volumes_checkbox.setChecked(
        True
    )
    segmentation_widget_with_data_atlas_space.region_seg.summarise_volumes_checkbox.setChecked(
        True
    )
    segmentation_widget_with_data_atlas_space.region_seg.run_region_analysis(
        override=True
    )

    # ensure data is saved before it is loaded again
    sleep(8)
    check_analysis(test_regions_dir, validate_regions_dir)

    # check saving didn't happen (default)
    test_saved_region = Path(test_regions_dir / "test_region.tiff")
    assert test_saved_region.exists() is False


def test_region_analysis_with_save(
    segmentation_widget_with_data_atlas_space, test_regions_dir
):
    segmentation_widget_with_data_atlas_space.region_seg.calculate_volumes_checkbox.setChecked(
        True
    )
    segmentation_widget_with_data_atlas_space.region_seg.summarise_volumes_checkbox.setChecked(
        True
    )
    segmentation_widget_with_data_atlas_space.region_seg.save_checkbox.setChecked(
        True
    )
    segmentation_widget_with_data_atlas_space.region_seg.run_region_analysis(
        override=True
    )

    # ensure data is saved before it is loaded again
    sleep(8)
    check_analysis(test_regions_dir, validate_regions_dir)
    check_saving(test_regions_dir, validate_regions_dir)


def test_region_save(
    segmentation_widget_with_data_atlas_space, test_regions_dir
):
    segmentation_widget_with_data_atlas_space.save(override=True)
    # ensure data is saved before it is loaded again
    sleep(8)
    check_saving(test_regions_dir, validate_regions_dir)


def test_region_export(
    segmentation_widget_with_data_atlas_space, test_regions_dir
):
    segmentation_widget_with_data_atlas_space.export_to_brainrender(
        override=True
    )

    # ensure data is saved before it is loaded again
    sleep(8)
    cmp(
        validate_regions_dir / "test_region.obj",
        test_regions_dir / "test_region.obj",
    )


def check_analysis(test_regions_dir, validate_regions_dir):
    region_csv_validate = pd.read_csv(validate_regions_dir / "test_region.csv")
    region_csv_test = pd.read_csv(test_regions_dir / "test_region.csv")
    pd.testing.assert_frame_equal(region_csv_test, region_csv_validate)

    summary_csv_validate = pd.read_csv(validate_regions_dir / "summary.csv")
    summary_csv_test = pd.read_csv(test_regions_dir / "summary.csv")
    pd.testing.assert_frame_equal(summary_csv_test, summary_csv_validate)


def check_saving(test_regions_dir, validate_regions_dir):
    image_validate = imread(validate_regions_dir / "test_region.tiff")
    image_test = imread(test_regions_dir / "test_region.tiff")
    np.testing.assert_array_equal(image_test, image_validate)



from pathlib import Path
from time import sleep

import numpy as np
import pandas as pd
import pytest

brainreg_dir = Path.cwd() / "tests" / "data" / "brainreg_output"
validate_tracks_dir = brainreg_dir / "segmentation" / "atlas_space" / "tracks"


@pytest.fixture
def test_tracks_dir(tmp_path):
    tmp_input_dir = tmp_path / "brainreg_output"
    test_tracks_dir = tmp_input_dir / "segmentation" / "atlas_space" / "tracks"
    return test_tracks_dir


def test_track_widget_layer_numbers(
    segmentation_widget_with_data_atlas_space, test_tracks_dir
):
    assert len(segmentation_widget_with_data_atlas_space.viewer.layers) == 4
    assert len(segmentation_widget_with_data_atlas_space.track_layers) == 1


def test_add_new_track(
    segmentation_widget_with_data_atlas_space, test_tracks_dir
):
    segmentation_widget_with_data_atlas_space.track_seg.add_track()
    assert len(segmentation_widget_with_data_atlas_space.viewer.layers) == 5
    assert len(segmentation_widget_with_data_atlas_space.track_layers) == 2
    assert (
        segmentation_widget_with_data_atlas_space.track_layers[0].name
        == "test_track"
    )
    assert (
        segmentation_widget_with_data_atlas_space.track_layers[1].name
        == "track_1"
    )
    assert (
        len(segmentation_widget_with_data_atlas_space.track_layers[0].data)
        == 6
    )


def test_add_existing_track(
    segmentation_widget_with_data_atlas_space, test_tracks_dir
):
    segmentation_widget = segmentation_widget_with_data_atlas_space

    # importing existing track
    points = np.array([[0, 0, 0], [1, 1, 1], [2, 2, 2], [3, 3, 3]])
    test_layer = segmentation_widget.viewer.add_points(
        points, name="test_track2"
    )
    segmentation_widget.viewer.layers.selection.select_only(test_layer)
    segmentation_widget.track_seg.add_track_from_existing_layer(override=True)
    assert len(segmentation_widget.viewer.layers) == 5
    assert len(segmentation_widget.track_layers) == 2


def test_add_surface_point(
    segmentation_widget_with_data_atlas_space, test_tracks_dir
):
    segmentation_widget_with_data_atlas_space.track_seg.add_surface_points()
    assert (
        len(segmentation_widget_with_data_atlas_space.track_layers[0].data)
        == 7
    )


def test_track_analysis_without_save(
    segmentation_widget_with_data_atlas_space, test_tracks_dir
):
    segmentation_widget_with_data_atlas_space.track_seg.run_track_analysis(
        override=True
    )
    # check saving didn't happen (default)
    test_saved_track = Path(test_tracks_dir / "test_track.points")
    assert test_saved_track.exists() is False

    check_analysis(test_tracks_dir, validate_tracks_dir)


def test_track_analysis_with_save(
    segmentation_widget_with_data_atlas_space, test_tracks_dir, rtol=1e-10
):
    segmentation_widget_with_data_atlas_space.track_seg.save_checkbox.setChecked(
        True
    )
    segmentation_widget_with_data_atlas_space.track_seg.run_track_analysis(
        override=True
    )

    check_analysis(test_tracks_dir, validate_tracks_dir)
    check_saving(test_tracks_dir, validate_tracks_dir, rtol)


def test_track_save(
    segmentation_widget_with_data_atlas_space, test_tracks_dir, rtol=1e-10
):
    segmentation_widget_with_data_atlas_space.save(override=True)

    # ensure data is saved before it is loaded again
    sleep(8)
    check_saving(test_tracks_dir, validate_tracks_dir, rtol)


def test_track_export(
    segmentation_widget_with_data_atlas_space, test_tracks_dir
):
    segmentation_widget_with_data_atlas_space.track_seg.run_track_analysis(
        override=True
    )
    segmentation_widget_with_data_atlas_space.export_to_brainrender(
        override=True
    )

    # ensure data is saved before it is loaded again
    sleep(8)
    spline_validate = np.load(validate_tracks_dir / "test_track.npy")
    spline_test = np.load(test_tracks_dir / "test_track.npy")
    np.testing.assert_allclose(spline_validate, spline_test)


def compare_dataframes(df1, df2, threshold=0.9):
    """
    Function to check how many entries are the same,
    not how similar they are.

    Required due to slight differences between operating
    systems & architectures
    """
    if df1.shape != df2.shape:
        raise ValueError("DataFrames are not the same shape.")

    total_entries = df1.shape[0] * df1.shape[1]
    matching_entries = (df1 == df2).sum().sum()
    similarity = matching_entries / total_entries
    return similarity >= threshold


def check_analysis(test_tracks_dir, validate_tracks_dir):
    regions_validate = pd.read_csv(validate_tracks_dir / "test_track.csv")
    regions_test = pd.read_csv(test_tracks_dir / "test_track.csv")
    # Threshold of 0.9 is not ideal but sight differences in the analysis
    # can have big effects on the final result (e.g. assigning to
    # neighbouring anatomical areas
    assert compare_dataframes(regions_validate, regions_test)


def check_saving(test_tracks_dir, validate_tracks_dir, rtol):
    points_validate = pd.read_hdf(validate_tracks_dir / "test_track.points")
    points_test = pd.read_hdf(test_tracks_dir / "test_track.points")
    np.testing.assert_allclose(points_validate, points_test, rtol=rtol)
